from typing import TYPE_CHECKING

from compas_ifc.conversions.unit import IfcCompoundPlaneAngleMeasure_to_degrees

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcSite
else:
    IfcSite = object


class IfcSite(IfcSite):
    """Extension class for :class:`IfcSite`.

    Attributes
    ----------
    buildings : list[:class:`IfcBuilding`]
        The buildings of the site.
    building_elements : list[:class:`IfcBuildingElement`]
        The building elements of the site.
    geographic_elements : list[:class:`IfcGeographicElement`]
        The geographic elements of the site.
    location : tuple[float, float]
        The location of the site. In degrees (latitude, longitude).

    """

    @property
    def buildings(self):
        return self.children_by_type("IfcBuilding", recursive=True)

    @property
    def building_elements(self):
        return self.children_by_type("IfcBuildingElement", recursive=True)

    @property
    def geographic_elements(self):
        return self.children_by_type("IfcGeographicElement", recursive=True)

    @property
    def location(self):
        if self.RefLatitude and self.RefLongitude:
            return IfcCompoundPlaneAngleMeasure_to_degrees(self.RefLatitude), IfcCompoundPlaneAngleMeasure_to_degrees(self.RefLongitude)
        else:
            return None



from ifcopenshell.util.element import get_psets

from compas_ifc.entities.base import Base
from compas_ifc.model import Model

PRIMARY_MEASURE_TYPES = {
    str: "IfcLabel",
    float: "IfcReal",
    bool: "IfcBoolean",
    int: "IfcInteger",
}


def from_dict_to_ifc_properties(model: Model, properties: dict) -> list[Base]:
    """Convert a dictionary to a list of IfcProperties"""

    ifc_properties = []

    for key, value in properties.items():
        if isinstance(value, dict):
            subproperties = from_dict_to_ifc_properties(model, value)
            ifc_property = model.create("IfcComplexProperty", Name=key, UsageName="{}", HasProperties=subproperties)
            ifc_properties.append(ifc_property)

        elif isinstance(value, list):
            subproperties = from_dict_to_ifc_properties(model, {str(k): v for k, v in enumerate(value)})
            ifc_property = model.create("IfcComplexProperty", Name=key, UsageName="[]", HasProperties=subproperties)
            ifc_properties.append(ifc_property)

        elif isinstance(value, (str, float, bool, int)):
            nominal_value = model.create_value(value)
            ifc_property = model.create("IfcPropertySingleValue", Name=key, NominalValue=nominal_value)
            ifc_properties.append(ifc_property)
        else:
            raise ValueError(f"Unsupported value type: {type(value)}")

    return ifc_properties


def from_dict_to_pset(model: Model, properties: dict, name: str = None) -> Base:
    ifc_properties = from_dict_to_ifc_properties(model, properties)
    pset = model.create("IfcPropertySet", Name=name, HasProperties=ifc_properties)
    return pset


def from_psets_to_dict(element: Base) -> dict:
    psets = get_psets(element.entity, psets_only=True)

    def _convert_property(property):
        if isinstance(property, dict):
            if property.get("UsageName", None) == "[]":
                return [_convert_property(value) for value in property["properties"].values()]
            elif property.get("UsageName", None) == "{}":
                return {key: _convert_property(value) for key, value in property["properties"].items()}
            else:
                if "id" in property:
                    del property["id"]
                return {key: _convert_property(value) for key, value in property.items()}
        else:
            return property

    return _convert_property(psets)



from compas.datastructures import Mesh

from compas_ifc.entities.base import Base
from compas_ifc.model import Model


def mesh_to_IfcPolygonalFaceSet(model: Model, mesh: Mesh) -> Base:
    """
    Convert a COMPAS mesh to an IFC PolygonalFaceSet.
    """
    keys = sorted(mesh.vertices())
    vertices = []
    for key in keys:
        coords = mesh.vertex_coordinates(key)
        vertices.append((float(coords[0]), float(coords[1]), float(coords[2])))

    faces = []
    for fkey in mesh.faces():
        indexes = [keys.index(i) + 1 for i in mesh.face_vertices(fkey)]
        faces.append(model.create("IfcIndexedPolygonalFace", CoordIndex=indexes))

    return model.create(
        "IfcPolygonalFaceSet",
        Closed=mesh.is_closed(),
        Coordinates=model.create("IfcCartesianPointList3D", CoordList=vertices),
        Faces=faces,
    )


def mesh_to_IfcFaceBasedSurfaceModel(model: Model, mesh: Mesh) -> Base:
    """
    Convert a COMPAS mesh to an IFC FaceBasedSurfaceModel.
    """
    vertices = {}
    for key in mesh.vertices():
        coords = mesh.vertex_coordinates(key)
        vertex = model.create("IfcCartesianPoint", Coordinates=(float(coords[0]), float(coords[1]), float(coords[2])))
        vertices[key] = vertex

    faces = []
    for fkey in mesh.faces():
        indexes = [vertices[key] for key in mesh.face_vertices(fkey)]
        polyloop = model.create("IfcPolyLoop", Polygon=indexes)
        bound = model.create("IfcFaceOuterBound", Bound=polyloop, Orientation=True)
        face = model.create("IfcFace", Bounds=[bound])
        faces.append(face)

    face_set = model.create("IfcConnectedFaceSet", CfsFaces=faces)
    ifc_face_based_surface_model = model.create("IfcFaceBasedSurfaceModel", FbsmFaces=[face_set])

    return ifc_face_based_surface_model



from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import Point
from compas.geometry import Vector

from compas_ifc.entities.base import Base
from compas_ifc.model import Model


def IfcCartesianPoint_to_point(cartesian_point: Base) -> Point:
    """
    Convert an IFC CartesianPoint [cartesianpoint]_ to a COMPAS point.

    """
    return Point(*cartesian_point.Coordinates)


def IfcDirection_to_vector(direction: Base) -> Vector:
    """
    Convert an IFC Direction [direction]_ to a COMPAS vector.

    """
    return Vector(*direction.DirectionRatios)


def IfcVector_to_vector(vector: Base) -> Vector:
    """
    Convert an IFC Vector [vector]_ to a COMPAS vector.

    """
    direction = IfcDirection_to_vector(vector.Orientation)
    direction.scale(vector.Magnitude)
    return direction


def IfcLine_to_line(line: Base) -> Line:
    """
    Convert an IFC Line [line]_ to a COMPAS line.

    """
    point = IfcCartesianPoint_to_point(line.Pnt)
    vector = IfcDirection_to_vector(line.Dir)
    return Line(point, point + vector)


def IfcPlane_to_plane(plane: Base) -> Plane:
    """
    Convert an IFC Plane [plane]_ to a COMPAS plane.

    """
    point = IfcCartesianPoint_to_point(plane.Position.Location)
    normal = IfcDirection_to_vector(plane.Position.P[3])
    return Plane(point, normal)


def point_to_IfcCartesianPoint(model: Model, point: Point) -> Base:
    """
    Convert a COMPAS point to an IFC CartesianPoint.
    """
    return model.create("IfcCartesianPoint", Coordinates=(float(point.x), float(point.y), float(point.z)))


def vector_to_IfcDirection(model: Model, vector: Vector) -> Base:
    """
    Convert a COMPAS vector to an IFC Direction.
    """
    return model.create("IfcDirection", DirectionRatios=(float(vector.x), float(vector.y), float(vector.z)))


def frame_to_IfcAxis2Placement3D(model: Model, frame: Frame) -> Base:
    """
    Convert a COMPAS frame to an IFC Axis2Placement3D.
    """
    return model.create(
        "IfcAxis2Placement3D",
        Location=point_to_IfcCartesianPoint(model, frame.point),
        Axis=vector_to_IfcDirection(model, frame.zaxis),
        RefDirection=vector_to_IfcDirection(model, frame.xaxis),
    )


def frame_to_IfcPlane(model: Model, frame: Frame) -> Base:
    """
    Convert a COMPAS frame to an IFC Plane.
    """
    return model.create("IfcPlane", Position=frame_to_IfcAxis2Placement3D(model, frame))



def IfcCompoundPlaneAngleMeasure_to_degrees(angle_components):
    if len(angle_components) != 4:
        raise ValueError("Input must be a list or tuple of four elements: [degrees, minutes, seconds, millionths_of_second]")

    degrees, minutes, seconds, millionths = angle_components

    # Determine the sign based on the degrees component
    sign = -1 if degrees < 0 else 1

    # Use absolute values to avoid negative components in calculation
    degrees = abs(degrees)
    minutes = abs(minutes)
    seconds = abs(seconds)
    millionths = abs(millionths)

    # Convert millionths of a second to fractional seconds
    fractional_seconds = millionths / 1_000_000

    # Total seconds including the fractional part
    total_seconds = seconds + fractional_seconds

    # Convert everything to decimal degrees
    decimal_degrees = sign * (degrees + (minutes / 60) + (total_seconds / 3600))

    return decimal_degrees



from functools import reduce
from operator import mul

from compas.geometry import Frame
from compas.geometry import Point
from compas.geometry import Transformation
from compas.geometry import Vector

from compas_ifc.conversions.primitives import IfcCartesianPoint_to_point
from compas_ifc.conversions.primitives import IfcDirection_to_vector
from compas_ifc.entities.base import Base
from compas_ifc.model import Model


def create_IfcAxis2Placement3D(model: Model, point: Point = None, dir1: Vector = None, dir2: Vector = None) -> Base:
    """
    Create an IFC Axis2Placement3D from a point, a direction and a second direction.
    """
    point = model.create("IfcCartesianPoint", Coordinates=point or [0.0, 0.0, 0.0])
    dir1 = model.create("IfcDirection", DirectionRatios=dir1 or [0.0, 0.0, 1.0])
    dir2 = model.create("IfcDirection", DirectionRatios=dir2 or [1.0, 0.0, 0.0])
    axis2placement = model.create("IfcAxis2Placement3D", Location=point, Axis=dir1, RefDirection=dir2)
    return axis2placement


def frame_to_ifc_axis2_placement_3d(model: Model, frame: Frame) -> Base:
    return create_IfcAxis2Placement3D(model, point=frame.point, dir1=frame.zaxis, dir2=frame.xaxis)


def assign_entity_frame(entity: Base, frame: Frame):
    local_placement = frame_to_ifc_axis2_placement_3d(entity.model, frame)
    placement = entity.model.create("IfcLocalPlacement", RelativePlacement=local_placement)
    entity.ObjectPlacement = placement


def IfcLocalPlacement_to_transformation(placement: Base, scale: float = 1) -> Transformation:
    """
    Convert an IFC LocalPlacement [localplacement]_ to a COMPAS transformation.
    This will resolve all relative placements into one transformation wrt the global coordinate system.

    """
    stack = []
    while True:
        Location = placement.RelativePlacement.Location
        Axis = placement.RelativePlacement.Axis
        RefDirection = placement.RelativePlacement.RefDirection

        if Axis and RefDirection:
            zaxis = Vector(*Axis.DirectionRatios)
            xaxis = Vector(*RefDirection.DirectionRatios)
            yaxis = zaxis.cross(xaxis)
            xaxis = yaxis.cross(zaxis)
        else:
            xaxis = Vector.Xaxis()
            yaxis = Vector.Yaxis()

        point = Point(*Location.Coordinates) * scale
        frame = Frame(point, xaxis, yaxis)
        stack.append(frame)

        if not placement.PlacementRelTo:
            break

        placement = placement.PlacementRelTo

    matrices = [Transformation.from_frame(f) for f in stack]
    return reduce(mul, matrices[::-1])


def IfcLocalPlacement_to_frame(placement: Base) -> Frame:
    """
    Convert an IFC LocalPlacement to a COMPAS frame.
    """

    Location = placement.RelativePlacement.Location
    Axis = placement.RelativePlacement.Axis
    RefDirection = placement.RelativePlacement.RefDirection

    if Axis and RefDirection:
        zaxis = Vector(*Axis.DirectionRatios)
        xaxis = Vector(*RefDirection.DirectionRatios)
        yaxis = zaxis.cross(xaxis)
        xaxis = yaxis.cross(zaxis)
    else:
        xaxis = Vector.Xaxis()
        yaxis = Vector.Yaxis()

    point = Point(*Location.Coordinates)
    return Frame(point, xaxis, yaxis)


def IfcGridPlacement_to_transformation(placement: Base) -> Transformation:
    pass


def IfcAxis2Placement2D_to_frame(placement: Base) -> Frame:
    """
    Convert an IFC Axis2Placement2D [axis2placement2d]_ to a COMPAS frame.

    An Axis2Placement2D is a 2D placement based on a frame defined by a point and 2 vectors.

    """
    # use the coordinate system of the representation context to replace missing axes
    # use defaults if also those not available
    point = IfcCartesianPoint_to_point(placement.Location)
    zaxis = Vector.Zaxis()
    if placement.RefDirection:
        xaxis = IfcDirection_to_vector(placement.RefDirection)
    else:
        xaxis = Vector.Xaxis()
    yaxis = zaxis.cross(xaxis)
    return Frame(point, xaxis, yaxis)


def IfcAxis2Placement3D_to_frame(placement: Base) -> Frame:
    """
    Convert an IFC Axis2Placement3D [axis2placement3d]_ to a COMPAS frame.

    An Axis2Placement3D is a 3D placement based on a frame defined by a point and 2 vectors.

    """
    # use the coordinate system of the representation context to replace missing axes
    # use defaults if also those not available
    point = IfcCartesianPoint_to_point(placement.Location)
    if placement.Axis:
        zaxis = IfcDirection_to_vector(placement.Axis)
    else:
        zaxis = Vector.Zaxis()
    if placement.RefDirection:
        xaxis = IfcDirection_to_vector(placement.RefDirection)
    else:
        xaxis = Vector.Xaxis()
    yaxis = zaxis.cross(xaxis)
    xaxis = yaxis.cross(zaxis)
    return Frame(point, xaxis, yaxis)



"""
This module contains functions for converting geometry representations between COMPAS and IFC.
"""

from typing import Union

from compas.datastructures import Mesh
from compas.geometry import Box
from compas.geometry import Brep
from compas.geometry import Cone
from compas.geometry import Cylinder
from compas.geometry import Shape
from compas.geometry import Sphere

from compas_ifc.conversions.brep import brep_to_IfcAdvancedBrep
from compas_ifc.conversions.mesh import mesh_to_IfcFaceBasedSurfaceModel
from compas_ifc.conversions.shapes import box_to_IfcBlock
from compas_ifc.conversions.shapes import cone_to_IfcRightCircularCone
from compas_ifc.conversions.shapes import cylinder_to_IfcRightCircularCylinder
from compas_ifc.conversions.shapes import sphere_to_IfcSphere
from compas_ifc.entities.extensions import IfcProduct
from compas_ifc.model import Model

REPRESENTATION_CACHE = {}


def assign_body_representation(entity: IfcProduct, representation: Union[Shape, Mesh, Brep]):
    """
    Assign a representation to an entity.
    """

    model: Model = entity.model

    if id(representation) in REPRESENTATION_CACHE:
        entity.Representation = REPRESENTATION_CACHE[id(representation)]
        return

    # Convert COMPAS geometries to IFC corresponding representation
    if isinstance(representation, Shape):
        if isinstance(representation, Box):
            ifc_csg_primitive3d = box_to_IfcBlock(model, representation)
        elif isinstance(representation, Sphere):
            ifc_csg_primitive3d = sphere_to_IfcSphere(model, representation)
        elif isinstance(representation, Cone):
            ifc_csg_primitive3d = cone_to_IfcRightCircularCone(model, representation)
        elif isinstance(representation, Cylinder):
            ifc_csg_primitive3d = cylinder_to_IfcRightCircularCylinder(model, representation)
        else:
            raise NotImplementedError(f"Conversion of {type(representation)} to IFC not implemented.")

        ifc_csg_solid = model.create("IfcCsgSolid", TreeRootExpression=ifc_csg_primitive3d)

        items = [ifc_csg_solid]
        representation_type = "CSG"

    elif isinstance(representation, Mesh):
        ifc_representation = mesh_to_IfcFaceBasedSurfaceModel(model, representation)
        representation_type = "SurfaceModel"
        items = [ifc_representation]

    elif isinstance(representation, Brep):
        if model.file.use_occ:
            try:
                items = brep_to_IfcAdvancedBrep(model, representation)
                representation_type = "SolidModel"
            except Exception as e:
                print(f"WARNING BREP conversion failed: {e}")
                items = []
                representation_type = "SurfaceModel"

        else:
            mesh, _ = representation.to_tesselation()
            ifc_representation = mesh_to_IfcFaceBasedSurfaceModel(model, mesh)
            representation_type = "SurfaceModel"
            items = [ifc_representation]

    else:
        raise NotImplementedError(f"Conversion of {type(representation)} to IFC not implemented.")

    # QUESTION: When using OCCBrep from Extrusion, can we still keep the extrusion data?

    ifc_shape_representation = model.create(
        "IfcShapeRepresentation",
        ContextOfItems=model.file.default_body_context,
        RepresentationIdentifier="Body",
        RepresentationType=representation_type,
        Items=items,
    )

    ifc_product_definition_shape = model.create(
        "IfcProductDefinitionShape",
        Representations=[ifc_shape_representation],
    )

    entity.Representation = ifc_product_definition_shape
    REPRESENTATION_CACHE[id(representation)] = ifc_product_definition_shape

    # TODO: should not overwrite all property sets here
    # TODO: alternative 1: restructure the metadata, remove duplicated info like vertices
    # TODO: alternative 2: save data as compact json string
    # entity.property_sets = {
    #     "Pset_COMPAS": {
    #         "representation_id": ifc_product_definition_shape.id(),
    #         "compas_data": json.loads(representation.to_jsonstring()),
    #     }
    # }


def read_representation(model: Model, entity: IfcProduct):
    pass


if __name__ == "__main__":
    import compas
    from compas.geometry import Frame

    model = Model.template(schema="IFC2X3", unit="m")

    # geometry = Box.from_width_height_depth(1, 1, 1)
    geometry = Mesh.from_ply(compas.get("bunny.ply"))
    # geometry = Mesh.from_meshgrid(5, 2, 5, 2)

    product = model.create(geometry=geometry, parent=model.building_storeys[0], name="test", frame=Frame.worldXY())

    model.show()

    model.save("temp/representations/test.ifc")



from typing import TYPE_CHECKING

from ifcopenshell.util.element import get_psets

from compas_ifc.conversions.pset import from_dict_to_pset
from compas_ifc.conversions.pset import from_psets_to_dict

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcObject
else:
    IfcObject = object


class IfcObject(IfcObject):
    """Extension class for :class:`IfcObject`.

    Attributes
    ----------
    properties : dict
        The property sets of the object.
    quantities : dict
        The quantity sets of the object.
    """

    _psetsmap = {}

    @property
    def psetsmap(self):
        if id(self.file) not in self._psetsmap:
            self._psetsmap[id(self.file)] = {}
        return self._psetsmap[id(self.file)]

    @property
    def property_sets(self):
        return from_psets_to_dict(self)

    @property_sets.setter
    def property_sets(self, psets):
        for name, pset in psets.items():
            if id(pset) in self.psetsmap:
                ifc_property_set = self.psetsmap[id(pset)]
            else:
                ifc_property_set = from_dict_to_pset(self.file, pset, name)
                self.psetsmap[id(pset)] = ifc_property_set
                # TODO: remove unused psets

            self.file.create(
                "IfcRelDefinesByProperties",
                OwnerHistory=self.file.default_owner_history,
                RelatingPropertyDefinition=ifc_property_set,
                RelatedObjects=[self],
            )

    @property
    def quantity_sets(self):
        qtos = get_psets(self.entity, qtos_only=True)
        for qto in qtos.values():
            del qto["id"]
        return qtos



import difflib
import importlib
import multiprocessing
import os
import time
from typing import Any
from typing import Dict
from typing import Type
from typing import Union

import ifcopenshell
import numpy as np
from compas.geometry import Transformation
from ifcopenshell.api import run

import compas_ifc
from compas_ifc.brep import TessellatedBrep
from compas_ifc.entities.base import Base


class IFCFile(object):
    """The IFCFile class is a wrapper around an ifcopenshell file object. It provides low-level access to the IFC data.

    Attributes
    ----------
    filepath : str, optional
        The path to the IFC file.
    model : :class:`compas_ifc.model.Model`
        The model object.
    use_occ : bool
        Whether to use OCC for geometry processing.
    load_geometries : bool
        Whether to load the geometries of the IFC file.
    verbose : bool
        Whether to print verbose output.
    extensions : dict, optional
        A dictionary of custom extensions to be used with the IFC file.
    schema : :class:`ifcopenshell.schema.Schema`
        The IFC schema object.
    schema_name : str
        The name of the IFC schema.
    classes : list[:class:`compas_ifc.entities.base.Base`]
        A list of all the classes for this schema version.
    default_project : :class:`compas_ifc.entities.generated.IFC4.IfcProject`
        The default project in this file. Will be created if it does not exist.
    default_units : :class:`compas_ifc.entities.generated.IFC4.IfcUnitAssignment`
        The default units in this file. Will be created if it does not exist.
    default_owner_history : :class:`compas_ifc.entities.generated.IFC4.IfcOwnerHistory`
        The default owner history in this file. Will be created if it does not exist.
    default_context : :class:`compas_ifc.entities.generated.IFC4.IfcContext`
        The default context in this file. Will be created if it does not exist.
    default_body_context : :class:`compas_ifc.entities.generated.IFC4.IfcContext`
        The default body context in this file. Will be created if it does not exist.

    """

    def __init__(
        self, model, filepath: str = None, schema: str = "IFC4", use_occ: bool = False, load_geometries: bool = True, verbose: bool = True, extensions: Dict[str, Type] = None
    ):
        """
        Construct the IFCFile object.

        Parameters
        ----------
        model : :class:`compas_ifc.model.Model`
            The model object.
        filepath : str, optional
            The path to the IFC file. If not provided, a new IFC file is created.
        schema : str, optional
            The IFC schema to use. Default is "IFC4".
        use_occ : bool, optional
            Whether to use OCC for geometry processing. Default is False.
        load_geometries : bool, optional
            Whether to load the geometries of the IFC file. Default is True.
        verbose : bool, optional
            Whether to print verbose output. Default is True.
        extensions : Dict[str, Type]
            A dictionary of extensions to use, with the key being the IFC class name and the value being the extension class.

        """

        self.extensions = extensions
        self.verbose = verbose
        self.ensure_classes_generated()
        self._entitymap = {}
        self._geometrymap = {}
        self._stylemap = {}
        self._relationmap_aggregates = {}  # map of IfcRelAggregates
        self._relationmap_contains = {}  # map of IfcRelContainedInSpatialStructure
        self._default_context = None
        self._default_body_context = None
        self._default_units = None
        self._default_owner_history = None
        self._default_project = None
        self._classes = None

        self.filepath = filepath
        self.model = model
        self.use_occ = use_occ
        if filepath is None:
            self._file = ifcopenshell.file(schema=schema)
            self._file.wrapped_data.header.file_name.author = ["Unknown Author"]
            self._file.wrapped_data.header.file_name.organization = ["Unknown Organization"]
            if self.verbose:
                print("IFC file created in schema: {}".format(schema))
        else:
            self._file = ifcopenshell.open(filepath)
            if self.verbose:
                print("IFC file loaded: {}".format(filepath))

        self._schema = ifcopenshell.ifcopenshell_wrapper.schema_by_name(self._file.schema)

        if load_geometries and filepath is not None:
            self.load_geometries()

    @property
    def schema(self) -> ifcopenshell.ifcopenshell_wrapper.schema_definition:
        return self._schema

    @property
    def schema_name(self) -> str:
        return self.schema.name()

    @property
    def classes(self) -> list[Base]:
        if not self._classes:
            module = importlib.import_module(f"compas_ifc.entities.generated.{self.schema_name}")
            classes = [x for x in dir(module) if x.startswith("Ifc") and x != "IfcRoot"]
            self._classes = classes
        return self._classes

    def ensure_classes_generated(self):
        """Check if the IFC classes are generated and generate them if not."""
        try:
            from compas_ifc.entities.generated import IFC2X3  # noqa: F401
            from compas_ifc.entities.generated import IFC4  # noqa: F401
        except ImportError:
            if self.verbose:
                print("IFC classes not found. Generating classes...")
            from compas_ifc.entities.generator import Generator

            generator = Generator(schema="IFC2X3")
            generator.generate()

            generator = Generator(schema="IFC4")
            generator.generate()
            if self.verbose:
                print("IFC classes generated.\n\n")

    def file_size(self) -> float:
        """Get the size of the IFC file in MB."""
        file_stats = os.stat(self.filepath)
        size_in_mb = file_stats.st_size / (1024 * 1024)
        size_in_mb = round(size_in_mb, 2)
        return size_in_mb

    def from_entity(self, entity: ifcopenshell.entity_instance) -> Base:
        """
        Convert an ifcopenshell entity to a compas_ifc entity. Will return an existing entity if it has already been converted.

        Parameters
        ----------
        entity : :class:`ifcopenshell.entity_instance`
            The ifcopenshell entity to convert.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The converted compas_ifc entity.

        """
        if not isinstance(entity, ifcopenshell.entity_instance):
            raise TypeError("Input is not an ifcopenshell.entity_instance")

        _id = entity.id()

        if _id in self._entitymap and _id != 0:
            return self._entitymap[_id]
        else:
            entity = Base(entity, file=self, extensions=self.extensions)
            if _id != 0:
                self._entitymap[_id] = entity
            return entity

    def get_entities_by_type(self, type_name: str) -> list[Base]:
        """
        Get all entities of a given type.

        Parameters
        ----------
        type_name : str
            The name of the type to get entities of.

        Returns
        -------
        list[:class:`compas_ifc.entities.base.Base`]
            A list of all entities of the given type.

        """
        entities = self._file.by_type(type_name)
        return [self.from_entity(entity) for entity in entities]

    def get_entities_by_name(self, name: str) -> list[Base]:
        """
        Get all entities with a given name.

        Parameters
        ----------
        name : str
            The name to search for.

        Returns
        -------
        list[:class:`compas_ifc.entities.base.Base`]
            A list of all entities with the given name.
        """
        return [entity for entity in self.get_entities_by_type("IfcRoot") if entity.Name == name]

    def get_entity_by_global_id(self, global_id: str) -> Base:
        """
        Get an entity by its global ID.

        Parameters
        ----------
        global_id : str
            The global ID to search for.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The entity with the given global ID.
        """
        return self.from_entity(self._file.by_guid(global_id))

    def get_entity_by_id(self, id: int) -> Base:
        """
        Get an entity by its file ID.

        Parameters
        ----------
        id : int
            The ID to search for.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The entity with the given ID.
        """
        return self.from_entity(self._file.by_id(id))

    def get_preloaded_geometry(self, entity: Base) -> "TessellatedBrep":
        """
        Get the preloaded geometry of an entity.

        Parameters
        ----------
        entity : :class:`compas_ifc.entities.base.Base`
            The entity to get the geometry of.

        Returns
        -------
        :class:`compas_ifc.brep.TessellatedBrep`
            The preloaded geometry of the entity. (OCCBrep if use_occ is True)
        """
        return self._geometrymap.get(entity.entity.id())

    def get_preloaded_style(self, entity: Base) -> dict:
        """
        Get the preloaded style of an entity.
        """
        return self._stylemap.get(entity.entity.id(), {})

    def load_geometries(self, include=None, exclude=None):
        """
        Load all the geometries of the IFC file using a fast multithreaded iterator.

        Parameters
        ----------
        include : list[str], optional
            A list of entity types to include.
        exclude : list[str], optional
            A list of entity types to exclude.

        """
        if self.verbose:
            print("Loading geometries...")
        import ifcopenshell.geom

        settings = ifcopenshell.geom.settings()
        settings.set(settings.CONVERT_BACK_UNITS, True)
        if self.use_occ:
            settings.set(settings.USE_PYTHON_OPENCASCADE, True)

        iterator = ifcopenshell.geom.iterator(settings, self._file, multiprocessing.cpu_count(), include=include, exclude=exclude)
        start = time.time()
        if iterator.initialize():
            while True:
                shape = iterator.get()
                if self.use_occ:
                    from compas_occ.brep import OCCBrep

                    brep = OCCBrep.from_shape(shape.geometry)

                    shellcolors = []
                    for style_id, style in zip(shape.style_ids, shape.styles):
                        if style_id == -1:
                            shellcolors.append((0.5, 0.5, 0.5, 1.0))
                        else:
                            shellcolors.append(style)

                    self._geometrymap[shape.data.id] = brep
                    self._stylemap[shape.data.id] = {"shellcolors": shellcolors}

                else:
                    from .brep import TessellatedBrep

                    matrix = shape.transformation.matrix.data
                    faces = shape.geometry.faces
                    edges = shape.geometry.edges
                    verts = shape.geometry.verts

                    matrix = np.array(matrix).reshape((4, 3))
                    matrix = np.hstack([matrix, np.array([[0], [0], [0], [1]])])
                    matrix = matrix.transpose()
                    transformation = Transformation.from_matrix(matrix.tolist())

                    facecolors = []
                    for m_id in shape.geometry.material_ids:
                        if m_id == -1:
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            facecolors.append([0.5, 0.5, 0.5, 1])
                            continue
                        material = shape.geometry.materials[m_id]
                        color = (*material.diffuse, 1 - material.transparency)
                        facecolors.append(color)
                        facecolors.append(color)
                        facecolors.append(color)

                    brep = TessellatedBrep(vertices=verts, edges=edges, faces=faces)

                    brep.transform(transformation)
                    self._geometrymap[shape.id] = brep
                    self._stylemap[shape.id] = {"facecolors": facecolors}

                if not iterator.next():
                    break

        if self.verbose:
            print(f"Time to load all {len(self._geometrymap)} geometries {(time.time() - start):.3f}s")

    def save(self, path: str):
        """
        Save the IFC file to a given path.
        """
        self._file.write(path)

    def export(self, path: str, entities: list[Base] = [], as_snippet: bool = False, export_materials: bool = True, export_properties: bool = True, export_styles: bool = True):
        """
        Export a subset of the IFC file to a new IFC file.

        Parameters
        ----------
        path : str
            The path to save the exported IFC file to.
        entities : list[:class:`compas_ifc.entities.base.Base`]
            The entities to export.
        as_snippet : bool
            Whether to export as a snippet, without the full spatial hierarchy. Default is False.
        export_materials : bool
            Whether to export materials. Default is True.
        export_properties : bool
            Whether to export properties. Default is True.
        export_styles : bool
            Whether to export styles. Default is True.

        """
        new_file = IFCFile(None, schema=self.schema_name)

        exported = {}

        def export_entity(entity: Union[Base, Any], file: IFCFile):
            if not isinstance(entity, Base):
                return entity

            if entity.is_a("IfcOwnerHistory"):
                return file.default_owner_history

            if entity in exported:
                return exported[entity]

            parent = getattr(entity, "parent", None)
            if parent and not as_snippet:
                new_parent_entity = export_entity(parent, file)

            new_entity = file._create_entity(entity.is_a())
            exported[entity] = new_entity

            if parent and not as_snippet:
                if hasattr(entity, "ContainedInStructure") and entity.ContainedInStructure():
                    file._create_entity("IfcRelContainedInSpatialStructure", RelatingStructure=new_parent_entity, RelatedElements=[new_entity])
                else:
                    file._create_entity("IfcRelAggregates", RelatingObject=new_parent_entity, RelatedObjects=[new_entity])

            for key, attr in entity.attributes.items():
                # Skip Representation and ObjectPlacement if not in the list of entities to export
                if key in ["Representation", "ObjectPlacement"] and entity not in entities:
                    continue

                if isinstance(attr, Base):
                    attr = export_entity(attr, file)
                elif isinstance(attr, (list, tuple)):
                    attr = [export_entity(a, file) for a in attr]

                setattr(new_entity, key, attr)

            if export_properties and hasattr(entity, "property_sets"):
                new_entity.property_sets = entity.property_sets

            if export_materials and hasattr(entity, "HasAssociations"):
                # TODO: create material settor on class extension.
                for relation in entity.HasAssociations():
                    if relation.is_a("IfcRelAssociatesMaterial"):
                        new_material_entity = export_entity(relation.RelatingMaterial, file)
                        file._create_entity("IfcRelAssociatesMaterial", RelatingMaterial=new_material_entity, RelatedObjects=[new_entity])

            if export_styles and hasattr(entity, "StyledByItem") and entity.entity.id():
                # TODO: create style settor on class extension.
                for style_item in entity.StyledByItem():
                    styles = [export_entity(s, file) for s in style_item.Styles]
                    file._create_entity("IfcStyledItem", Styles=styles, Item=new_entity)

            return new_entity

        for entity in entities:
            export_entity(entity, new_file)

        new_file.save(path)

    def create(self, cls=None, parent=None, geometry=None, frame=None, properties=None, **kwargs) -> Base:
        """
        Create an entity in this model.

        Parameters
        ----------
        cls : str
            The class of the entity to create. Defaults to an `IfcBuildingElementProxy`.
        parent : :class:`compas_ifc.entities.base.Base`
            The parent entity of the new entity.
        geometry : :class:`compas.geometry.Geometry` or :class:`compas.datastructures.Datastructure`
            The geometry of the new entity.
        frame : :class:`compas.geometry.Frame`
            The placement frame of the new entity.
        properties : dict
            The custom property sets to be added to the new entity.
        **kwargs
            Additional keyword arguments to be added to the new entity.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The newly created entity.

        """
        if isinstance(cls, type):
            cls_name = cls.__name__
        else:
            cls_name = cls

        if cls_name is None:
            cls_name = "IfcBuildingElementProxy"

        if cls_name not in self.classes:
            matched_classes = self.search_ifc_classes(cls_name)
            if not matched_classes:
                raise ValueError(f"Class {cls_name} not found.")
            else:
                cls_name = matched_classes[0]

        entity = self._create_entity(cls_name, **kwargs)

        if parent:
            self.create_relationship(parent, entity)

        if geometry:
            # TODO: Deal with instancing
            entity.geometry = geometry

        if frame:
            entity.frame = frame

        if properties:
            entity.property_sets = properties

        if entity.is_a("IfcRoot"):
            entity.GlobalId = ifcopenshell.guid.new()

        return entity

    def create_relationship(self, parent: Base, child: Base) -> Base:
        """
        Create the correct relationship between two entities based on their types.

        Parameters
        ----------
        parent : :class:`compas_ifc.entities.base.Base`
            The parent entity.
        child : :class:`compas_ifc.entities.base.Base`
            The child entity.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The created relationship.

        """

        guid = ifcopenshell.guid.new()
        if parent.is_a("IfcSpatialStructureElement"):
            if child.is_a("IfcSpatialStructureElement"):
                return self._create_entity("IfcRelAggregates", GlobalId=guid, RelatingObject=parent, RelatedObjects=[child])
            return self._create_entity("IfcRelContainedInSpatialStructure", GlobalId=guid, RelatingStructure=parent, RelatedElements=[child])

        if parent.is_a("IfcElementAssembly") or child.is_a("IfcBuildingElementPart"):
            return self._create_entity("IfcRelAggregates", GlobalId=guid, RelatingObject=parent, RelatedObjects=[child])

        if parent.is_a("IfcPort"):
            if not child.is_a("IfcPort"):
                return self._create_entity("IfcRelConnectsPortToElement", GlobalId=guid, RelatingPort=parent, RelatedElement=child)
            return self._create_entity("IfcRelConnectsPorts", GlobalId=guid, RelatingPort=parent, RelatedPort=child)

        if parent.is_a("IfcElement") and child.is_a("IfcElement"):
            return self._create_entity("IfcRelConnectsElements", GlobalId=guid, RelatingElement=parent, RelatedElement=child)

        if parent.is_a("IfcGroup"):
            return self._create_entity("IfcRelAssignsToGroup", GlobalId=guid, RelatingGroup=parent, RelatedObjects=[child])

        # Default case
        return self._create_entity("IfcRelAggregates", GlobalId=guid, RelatingObject=parent, RelatedObjects=[child])

    def search_ifc_classes(self, name: str, n: int = 5) -> list[Type["Base"]]:
        """
        Search for IFC classes by name

        Parameters
        ----------
        name : str
            The name of the IFC class to search for.
        n : int
            The number of results to return. Default is 5.

        Returns
        -------
        list[Type[:class:`compas_ifc.entities.base.Base`]]
            A list of IFC classes that match the search query.

        """
        classes_dict = {c.lower(): c for c in self.classes}
        classes_lower = list(classes_dict.keys())
        closest_matches_lower = difflib.get_close_matches(name.lower(), classes_lower, n=n)
        closest_matches = [classes_dict[match] for match in closest_matches_lower]
        return closest_matches

    def remove(self, entity: Union[Base, list[Base]]):
        """
        Remove an entity from this model.

        Parameters
        ----------
        entity : :class:`compas_ifc.entities.base.Base` or list[:class:`compas_ifc.entities.base.Base`]
            The entity or entities to remove.

        """
        if isinstance(entity, Base):
            entity = [entity]

        print(f"Removing {len(entity)} entities...")
        ifcopenshell.util.element.batch_remove_deep2(self._file)
        for e in entity:
            ifcopenshell.util.element.remove_deep2(self._file, e.entity)
        self._file = ifcopenshell.util.element.unbatch_remove_deep2(self._file)
        print("Removal done.")

    def _create_entity(self, cls_name, **kwargs) -> Base:
        camel_case_kwargs = {}

        for key, value in kwargs.items():
            if isinstance(value, Base):
                kwargs[key] = value.entity
            elif isinstance(value, (list, tuple)):
                kwargs[key] = [v.entity if isinstance(v, Base) else v for v in value]

            # if key is a snake_case key, convert it to camelCase
            if key[0].isupper():
                camel_case_kwargs[key] = kwargs[key]
            else:
                camel_case_key = "".join([word.capitalize() for word in key.split("_")])
                camel_case_kwargs[camel_case_key] = kwargs[key]

        entity = self._file.create_entity(cls_name, **camel_case_kwargs)
        return self.from_entity(entity)

    def create_value(self, value):
        """
        Create corresponding IfcValue from a Python value.
        """

        PRIMARY_MEASURE_TYPES = {
            str: "IfcLabel",
            float: "IfcReal",
            bool: "IfcBoolean",
            int: "IfcInteger",
        }

        primary_measure_type = PRIMARY_MEASURE_TYPES[type(value)]
        ifc_value = self._file.create_entity(primary_measure_type, value)
        return ifc_value

    @property
    def default_project(self) -> Base:
        projects = self._file.by_type("IfcProject")
        if projects:
            self._default_project = self.from_entity(projects[0])
        else:
            self._default_project = self._create_entity("IfcProject", GlobalId=ifcopenshell.guid.new(), Name="Default Project")
            self.default_units
            self.default_body_context
            self.default_owner_history
            return self._default_project

        return self._default_project

    @property
    def default_units(self) -> Base:
        if not self._default_units:
            if self.default_project.UnitsInContext:
                self._default_units = self.default_project.UnitsInContext
            else:
                length_unit = self.create("IfcUnit", UnitType="LENGTHUNIT", Prefix="MILLI", Name="METRE")
                area_unit = self.create("IfcUnit", UnitType="AREAUNIT", Prefix="MILLI", Name="SQUARE_METRE")
                volume_unit = self.create("IfcUnit", UnitType="VOLUMEUNIT", Prefix="MILLI", Name="CUBIC_METRE")
                plane_angle_unit = self.create("IfcUnit", UnitType="PLANEANGLEUNIT", Name="RADIAN")
                unit_assignment = self.create("IfcUnitAssignment", Units=[length_unit, area_unit, volume_unit, plane_angle_unit])
                self.default_project.UnitsInContext = unit_assignment
                self._default_units = unit_assignment
        return self._default_units

    @property
    def default_context(self) -> Base:
        if not self._default_context:
            contexts = self.get_entities_by_type("IfcGeometricRepresentationContext")
            for context in contexts:
                if context.ContextType == "Model":
                    self._default_context = context
                    return self._default_context

            self._default_context = self.from_entity(run("context.add_context", self._file, context_type="Model"))
        return self._default_context

    @property
    def default_body_context(self) -> Base:
        if not self._default_body_context:
            contexts = self.get_entities_by_type("IfcGeometricRepresentationSubContext")
            for context in contexts:
                if context.ContextIdentifier == "Body":
                    self._default_body_context = context
                    return self._default_body_context

            self._default_body_context = self.from_entity(
                run(
                    "context.add_context",
                    self._file,
                    context_type="Model",
                    context_identifier="Body",
                    target_view="MODEL_VIEW",
                    parent=self.default_context.entity,
                )
            )

        return self._default_body_context

    @property
    def default_owner_history(self) -> Base:
        # We will create a new owner history since we are updating the file
        if not self._default_owner_history:
            person = self._create_entity("IfcPerson", FamilyName="Unknown Author")
            organization = self._create_entity("IfcOrganization", Name="compas.dev")
            person_and_org = self._create_entity("IfcPersonAndOrganization", ThePerson=person, TheOrganization=organization)
            application = self._create_entity(
                "IfcApplication",
                ApplicationDeveloper=organization,
                Version=compas_ifc.__version__,
                ApplicationFullName="compas_ifc",
                ApplicationIdentifier="compas_ifc v" + compas_ifc.__version__,
            )

            owner_history = self._create_entity(
                "IfcOwnerHistory",
                OwningUser=person_and_org,
                OwningApplication=application,
                CreationDate=int(time.time()),
            )

            self._default_owner_history = owner_history
        return self._default_owner_history



from typing import TYPE_CHECKING
from typing import Dict
from typing import Generator
from typing import Type
from typing import Union

from compas.data import Data
from compas.datastructures import Datastructure
from compas.geometry import Frame
from compas.geometry import Geometry
from compas.geometry import Transformation

from compas_ifc.file import IFCFile

if TYPE_CHECKING:
    import ifcopenshell.ifcopenshell_wrapper

    from compas_ifc.entities.base import Base
    from compas_ifc.entities.generated.IFC4 import IfcBuilding
    from compas_ifc.entities.generated.IFC4 import IfcBuildingElement
    from compas_ifc.entities.generated.IFC4 import IfcBuildingStorey
    from compas_ifc.entities.generated.IFC4 import IfcProject
    from compas_ifc.entities.generated.IFC4 import IfcSite


class Model(Data):
    """The Model class is the COMPAS IFC's main entry point.
    It is a intuitive abstraction on top of an IFC file, providing a set of easy-to-use methods for interacting with the IFC data.

    Attributes
    ----------
    file : :class:`compas_ifc.file.IFCFile`
        The IFC file object.
    schema : :class:`ifcopenshell.schema.Schema`
        The IFC schema object.
    schema_name : str
        The name of the IFC schema.
    entities : Generator[:class:`compas_ifc.entities.base.Base`, None, None]
        A generator of all entities in the IFC file.
    project : :class:`compas_ifc.entities.generated.IFC4.IfcProject`
        The project entity.
    sites : list[:class:`compas_ifc.entities.generated.IFC4.IfcSite`]
        A list of all site entities.
    buildings : list[:class:`compas_ifc.entities.generated.IFC4.IfcBuilding`]
        A list of all building entities.
    building_storeys : list[:class:`compas_ifc.entities.generated.IFC4.IfcBuildingStorey`]
        A list of all building storey entities.
    building_elements : list[:class:`compas_ifc.entities.generated.IFC4.IfcBuildingElement`]
        A list of all building element entities.
    unit : str
        The unit of the IFC file. Can be "mm", "cm", or "m".

    """

    def __init__(self, filepath: str = None, schema: str = "IFC4", use_occ: bool = False, load_geometries: bool = True, verbose: bool = True, extensions: Dict[str, Type] = None):
        """
        Construct the Model object.

        Parameters
        ----------
        filepath : str
            The path to the IFC file.
        schema : str
            The IFC schema to use. Default is "IFC4".
        use_occ : bool
            Whether to use OCC for geometry processing. Default is False.
        load_geometries : bool
            Whether to pre-load geometries from the IFC file using multi-threading. Default is True.
        verbose : bool
            Whether to print verbose output. Default is True.
        extensions : Dict[str, Type]
            A dictionary of extensions to use, with the key being the IFC class name and the value being the extension class.

        """
        self.file = IFCFile(self, filepath=filepath, schema=schema, use_occ=use_occ, load_geometries=load_geometries, verbose=verbose, extensions=extensions)

    @property
    def schema(self) -> "ifcopenshell.ifcopenshell_wrapper.schema_definition":
        return self.file.schema

    @property
    def schema_name(self) -> str:
        return self.file.schema_name

    @property
    def entities(self) -> Generator["Base", None, None]:
        for entity in self.file._file:
            yield self.file.from_entity(entity)

    @property
    def project(self) -> "IfcProject":
        projects = self.file.get_entities_by_type("IfcProject")
        return projects[0] if projects else None

    @property
    def sites(self) -> list["IfcSite"]:
        return self.file.get_entities_by_type("IfcSite")

    @property
    def buildings(self) -> list["IfcBuilding"]:
        return self.file.get_entities_by_type("IfcBuilding")

    @property
    def building_storeys(self) -> list["IfcBuildingStorey"]:
        return self.file.get_entities_by_type("IfcBuildingStorey")

    @property
    def building_elements(self) -> list["IfcBuildingElement"]:
        return self.file.get_entities_by_type("IfcBuildingElement")

    @property
    def unit(self) -> str:
        length_unit = self.project.length_unit
        if length_unit.Name == "METRE" and length_unit.Prefix == "MILLI":
            return "mm"
        elif length_unit.Name == "METRE" and length_unit.Prefix == "CENTI":
            return "cm"
        elif length_unit.Name == "METRE" and not length_unit.Prefix:
            return "m"

    @unit.setter
    def unit(self, value: str):
        if value == "mm":
            self.project.length_unit.Prefix = "MILLI"
        elif value == "cm":
            self.project.length_unit.Prefix = "CENTI"
        elif value == "m":
            self.project.length_unit.Prefix = None
        else:
            raise ValueError("Invalid unit. Use 'mm', 'cm', or 'm'.")

    def get_entities_by_type(self, type_name: str) -> list["Base"]:
        """Get entities by type name

        Parameters
        ----------
        type_name : str
            The name of the type to get entities for.

        Returns
        -------
        list[:class:`compas_ifc.entities.base.Base`]
            A list of entities of the given type.
        """
        return self.file.get_entities_by_type(type_name)

    def get_entities_by_name(self, name: str) -> list["Base"]:
        """Get entities by name

        Parameters
        ----------
        name : str
            The name of the entity to get.

        Returns
        -------
        list[:class:`compas_ifc.entities.base.Base`]
            A list of entities with the given name.
        """
        return self.file.get_entities_by_name(name)

    def get_entity_by_global_id(self, global_id: str) -> "Base":
        """Get an entity by global ID

        Parameters
        ----------
        global_id : str
            The global ID of the entity to get.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The entity with the given global ID.
        """
        return self.file.get_entity_by_global_id(global_id)

    def get_entity_by_id(self, id: int) -> "Base":
        """Get an entity by ID

        Parameters
        ----------
        id : int
            The ID of the entity to get.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The entity with the given ID.
        """
        return self.file.get_entity_by_id(id)

    def search_ifc_classes(self, name: str, n: int = 5) -> list[Type["Base"]]:
        """
        Search for IFC classes by name

        Parameters
        ----------
        name : str
            The name of the IFC class to search for.
        n : int
            The number of results to return. Default is 5.

        Returns
        -------
        list[Type[:class:`compas_ifc.entities.base.Base`]]
            A list of IFC classes that match the search query.

        """
        return self.file.search_ifc_classes(name, n)

    def print_summary(self):
        """Print a summary of the IFC file."""

        print("=" * 80)
        print("Schema: {}".format(self.schema_name))
        if self.file.filepath:
            print("File: {}".format(self.file.filepath))
            print("Size: {} MB".format(self.file.file_size()))
        if self.project:
            print("Project: {}".format(self.project.Name))
            print("Description: {}".format(self.project.Description))
        print("Number of sites: {}".format(len(self.sites)))
        print("Number of buildings: {}".format(len(self.buildings)))
        print("Number of building elements: {}".format(len(self.building_elements)))
        print("=" * 80)

    def print_spatial_hierarchy(self, max_depth: int = 3):
        """Print the spatial hierarchy of the IFC file."""
        self.project.print_spatial_hierarchy(max_depth=max_depth)

    def save(self, path: str):
        """Save the IFC file."""
        self.file.save(path)

    def export(self, path: str, entities: list["Base"] = [], as_snippet: bool = False, export_materials: bool = True, export_properties: bool = True, export_styles: bool = True):
        """
        Export a subset of the IFC file to a new IFC file.

        Parameters
        ----------
        path : str
            The path to save the exported IFC file to.
        entities : list[:class:`compas_ifc.entities.base.Base`]
            The entities to export.
        as_snippet : bool
            Whether to export as a snippet, without the full spatial hierarchy. Default is False.
        export_materials : bool
            Whether to export materials. Default is True.
        export_properties : bool
            Whether to export properties. Default is True.
        export_styles : bool
            Whether to export styles. Default is True.

        """
        self.file.export(path, entities=entities, as_snippet=as_snippet, export_materials=export_materials, export_properties=export_properties, export_styles=export_styles)

    def show(self, entity: "Base" = None, linear_deflection: float = 100):
        """Show the IFC file in a viewer, either the entire project or a single entity.

        Parameters
        ----------
        entity : :class:`compas_ifc.entities.base.Base`
            The entity to show. If None, the entire project will be shown.
        linear_deflection : float
            The linear deflection to use for the tesselation of BREP geometries.
            Should be adjusted based on the size of the model.
        """
        try:
            from compas_viewer import Viewer
            from compas_viewer.components import Treeform

        except ImportError:
            raise ImportError("The show method requires compas_viewer to be installed.")

        viewer = Viewer()
        print(f"Unit: {self.unit}")
        print(f"Using Linear Deflection: {linear_deflection}")
        viewer.unit = self.unit
        viewer.ui.sidebar.show_objectsetting = False

        entity_map = {}

        def parse_entity(entity, parent=None):
            obj = None
            name = f"[{entity.__class__.__name__}]{entity.Name}"
            transformation = Transformation.from_frame(entity.frame) if entity.frame else None
            if getattr(entity, "geometry", None) and not entity.is_a("IfcSpace"):
                obj = viewer.scene.add(entity.geometry, name=name, parent=parent, hide_coplanaredges=True, **entity.style, linear_deflection=linear_deflection)
                obj.transformation = transformation
            else:
                obj = viewer.scene.add_group(name=name, parent=parent)

            obj.attributes["entity"] = entity

            for child in entity.children:
                parse_entity(child, parent=obj)

            if obj:
                entity_map[id(obj)] = entity

        parse_entity(entity or self.project)

        treeform = Treeform()
        viewer.ui.sidebar.add(treeform)

        def update_treeform(form, node):
            entity = node.attributes["entity"]
            treeform.update_from_dict({"Attributes": entity.attributes, "PSets": getattr(entity, "property_sets", {})})

        viewer.ui.sidebar.sceneform.action = update_treeform

        viewer.show()

    def create(
        self, cls: str = "IfcBuildingElementProxy", parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs
    ) -> "Base":
        """
        Create an entity in this model.

        Parameters
        ----------
        cls : str
            The class of the entity to create. Defaults to an `IfcBuildingElementProxy`.
        parent : :class:`compas_ifc.entities.base.Base`
            The parent entity of the new entity.
        geometry : :class:`compas.geometry.Geometry` or :class:`compas.datastructures.Datastructure`
            The geometry of the new entity.
        frame : :class:`compas.geometry.Frame`
            The placement frame of the new entity.
        properties : dict
            The custom property sets to be added to the new entity.
        **kwargs
            Additional keyword arguments to be added to the new entity.

        Returns
        -------
        :class:`compas_ifc.entities.base.Base`
            The newly created entity.

        """
        return self.file.create(cls=cls, parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_value(self, value):
        return self.file.create_value(value)

    def create_default_project(self) -> "IfcProject":
        """Create a default project in this model if one does not exist."""
        return self.file.default_project

    def remove(self, entity: Union["Base", list["Base"]]):
        """
        Remove an entity from this model.

        Parameters
        ----------
        entity : :class:`compas_ifc.entities.base.Base` or list[:class:`compas_ifc.entities.base.Base`]
            The entity or entities to remove.

        """
        self.file.remove(entity)

    @classmethod
    def template(cls, schema: str = "IFC4", building_count: int = 1, storey_count: int = 1, unit: str = "mm", use_occ: bool = False) -> "Model":
        """Create a template model with a default project, site, building, and storey.

        Parameters
        ----------
        schema : str
            The IFC schema to use. Defaults to "IFC4".
        building_count : int
            The number of buildings to create. Defaults to 1.
        storey_count : int
            The number of storeys to create. Defaults to 1.
        unit : str
            The unit of the model. Defaults to "mm".

        Returns
        -------
        :class:`compas_ifc.model.Model`
            The newly created model.
        """
        model = cls(schema=schema, use_occ=use_occ)
        project = model.file.default_project
        site = model.create("IfcSite", parent=project, Name="Default Site")
        for i in range(building_count):
            building = model.create("IfcBuilding", parent=site, Name=f"Default Building {i+1}")
            for j in range(storey_count):
                model.create("IfcBuildingStorey", parent=building, Name=f"Default Storey {j+1}")

        model.unit = unit
        model.update_linear_deflection()
        return model

    # Helper functions to create specific IFC entities

    def create_wall(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a wall in this model."""
        return self.file.create(cls="IfcWall", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_slab(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a slab in this model."""
        return self.file.create(cls="IfcSlab", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_roof(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a roof in this model."""
        return self.file.create(cls="IfcRoof", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_window(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a window in this model."""
        return self.file.create(cls="IfcWindow", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_door(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a door in this model."""
        return self.file.create(cls="IfcDoor", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_column(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a column in this model."""
        return self.file.create(cls="IfcColumn", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_beam(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a beam in this model."""
        return self.file.create(cls="IfcBeam", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_railing(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a railing in this model."""
        return self.file.create(cls="IfcRailing", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)

    def create_stair(self, parent: "Base" = None, geometry: Union[Geometry, Datastructure] = None, frame: Frame = None, properties: dict = None, **kwargs) -> "Base":
        """Create a stair in this model."""
        return self.file.create(cls="IfcStair", parent=parent, geometry=geometry, frame=frame, properties=properties, **kwargs)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from compas_ifc.entities.generated.IFC4 import IfcElement
else:
    IfcElement = object


class IfcElement(IfcElement):
    """Extension class for :class:`IfcElement`.

    Attributes
    ----------
    parent : :class:`IfcElement`
        The parent element of the element.
    """

    @property
    def parent(self):
        relations = self.ContainedInStructure()
        if relations:
            return relations[0].RelatingStructure
        else:
            return super().parent



import numpy as np
from compas.datastructures import Mesh
from compas.geometry import Geometry
from compas.geometry import area_polygon
from compas.geometry import bounding_box
from compas.geometry import transform_points_numpy
from compas.geometry import volume_polyhedron


class TessellatedBrep(Geometry):
    def __init__(self, vertices=None, edges=None, faces=None, **kwargs):
        super().__init__(**kwargs)
        if vertices is None:
            vertices = []
        if edges is None:
            edges = []
        if faces is None:
            faces = []
        self.vertices = np.array(vertices).reshape(-1, 3)
        self.edges = np.array(edges).reshape(-1, 2)
        self.faces = np.array(faces).reshape(-1, 3)

    def transform(self, transformation):
        self.vertices = transform_points_numpy(self.vertices, transformation)

    def to_vertices_and_faces(self):
        return self.vertices, self.faces

    def to_mesh(self):
        mesh = Mesh.from_vertices_and_faces(self.vertices, self.faces)
        mesh.name = self.name
        return mesh

    @property
    def aabb(self):
        from compas.geometry import Box

        return Box.from_bounding_box(bounding_box(self.vertices))

    @property
    def obb(self):
        from compas.geometry import Box
        from compas.geometry import oriented_bounding_box_numpy

        return Box.from_bounding_box(oriented_bounding_box_numpy(self.vertices))

    @property
    def volume(self):
        """Calculate the volume of the tessellated BREP by converting to mesh.

        Returns
        -------
        float or None
            The volume of the BREP if the mesh is closed, None otherwise.
        """
        # Handle edge case: empty vertices or faces
        if len(self.vertices) == 0 or len(self.faces) == 0:
            return None

        try:
            mesh = self.to_mesh()

            # Check if mesh is closed - volume can only be calculated for closed meshes
            if not mesh.is_closed():
                return None

            # Get vertices and faces for volume calculation
            vertices, faces = mesh.to_vertices_and_faces()

            # Calculate volume using COMPAS geometry function
            return volume_polyhedron((vertices, faces))

        except Exception:
            # Return None if any error occurs during calculation
            return None

    @property
    def surface_area(self):
        """Calculate the surface area of the tessellated BREP by converting to mesh.

        Returns
        -------
        float or None
            The surface area of the BREP, or None if calculation fails.
        """
        # Handle edge case: empty vertices or faces
        if len(self.vertices) == 0 or len(self.faces) == 0:
            return None

        try:
            mesh = self.to_mesh()

            # Calculate total surface area by summing face areas
            total_area = 0.0
            for face_key in mesh.faces():
                # Get face vertices coordinates
                face_vertices = [mesh.vertex_coordinates(vertex) for vertex in mesh.face_vertices(face_key)]

                # Calculate area of this face using COMPAS geometry function
                face_area = area_polygon(face_vertices)
                total_area += face_area

            return total_area

        except Exception:
            # Return None if any error occurs during calculation
            return None
